require "othello_ruby/version"

module OthelloRuby
  EIGHT_DIRECTION = %w(left right upward under left_upward left_under right_upward right_under)
  REPLACE = {"a" => 0, "b" => 1, "c" => 2, "d" => 3, "e" => 4, "f" => 5, "g" => 6, "h" => 7}
  class Error < StandardError; end
  # Your code goes here...
  module Game
    module BordWork
      def get_reverse_pattern(active, passive, bit_string)
        patterns = []
        EIGHT_DIRECTION.each do |name|
          patterns << send("get_#{name}_pat", active, passive, bit_string)
        end
        patterns.delete(0)
        patterns.compact
      end

      EIGHT_DIRECTION.each do |name|
        define_method("get_#{name}_pat") do |active, passive, bit_string|
          return nil if ((active | passive) & bit_string) != 0
          pattern = 0
          mask = send("#{name}_direction", bit_string)
          while mask != 0 && (mask & passive) != 0
            pattern |= mask
            mask = send("#{name}_direction", mask)
          end
          if (mask & active) == 0
            return 0
          else
            return pattern
          end
        end
      end

      def left_direction(bit_string)
        (bit_string >> 1) & 0x7F7F7F7F7F7F7F7F
      end

      def right_direction(bit_string)
        (bit_string << 1) & 0xFEFEFEFEFEFEFEFE
      end

      def upward_direction(bit_string)
        bit_string >> 8
      end

      def under_direction(bit_string)
        (bit_string << 8) & 0xFFFFFFFFFFFFFFFF
      end

      def left_upward_direction(bit_string)
        bit_string >> 9
      end

      def left_under_direction(bit_string)
        (bit_string << 7) & 0xFFFFFFFFFFFFFFFF
      end

      def right_upward_direction(bit_string)
        bit_string >> 7
      end

      def right_under_direction(bit_string)
        (bit_string << 9) & 0xFFFFFFFFFFFFFFFF
      end
    end

    class Bord
      include BordWork
      attr_accessor :white, :black
      attr_reader :first_move_color, :passive_move_color
      def initialize(**options)
        @white_square = options[:white_square] || "0"
        @black_square = options[:black_square] || "1"
        @first_move_color    = options[:first_move_color]  || "black"
        @passive_move_color  = options[:passive_move_color]  || "white"
        @white = 0x0000001008000000
        @black = 0x0000000810000000
      end

      def show
        message = "    a   b   c   d   e   f   g   h\n  +---+---+---+---+---+---+---+---+\n"
        8.times do |row|
          message << "#{row+1} |"
          8.times do |col|
            if  @white[row * 8 + col] == 1
              message << " #{@white_square} "
            elsif @black[row * 8 + col] == 1
              message << " #{@black_square} "
            else
              message << "   "
            end
            message << "|"
          end
          message << "\n  +---+---+---+---+---+---+---+---+\n"
        end
        message
      end

      def white_count
        count = 0
        64.times do |i|
          count += 1 if @white[i] == 1
        end
        count
      end

      def black_count
        count = 0
        64.times do |i|
          count += 1 if @black[i] == 1
        end
        count
      end

      def white_set(coordinate)
        bit_string = convert_to_number(coordinate)
        patterns   = get_reverse_pattern(@white, @black, bit_string)
        if patterns.empty?
          nil
        else
          pattern = patterns.inject(&:+)
          @white ^= (bit_string | pattern)
          @black ^= pattern
        end
      end

      def black_set(coordinate)
        bit_string = convert_to_number(coordinate)
        patterns   = get_reverse_pattern(@black, @white, bit_string)
        if patterns.empty?
          nil
        else
          pattern = patterns.inject(&:|)
          @black ^= (bit_string | pattern)
          @white ^= pattern
        end
      end

      def search_white_position
        positions = []
        64.times do |num|
          positions << num unless  get_reverse_pattern(@white, @black, 1 << num).empty?
        end
        positions.map!{|position| 1 << position }
        positions.inject(&:|)
      end

      def search_black_position
        positions = []
        64.times do |num|
          positions << num unless get_reverse_pattern(@black, @white, 1 << num).empty?
        end
        positions.map{|position| 1 << position }
        positions.inject(&:|)
      end
    end

    class User
      include BordWork
      def initialize(color)
        @color = color
      end

      def set(bord, coordinate)
        bit_string = convert_to_number(coordinate)
        case @color
        when bord.first_move_color
          patterns = get_reverse_pattern(bord.black, bord.white, bit_string)
        when bord.passive_move_color
          patterns = get_reverse_pattern(bord.white, bord.black, bit_string)
        else
          patterns = []
        end
        if patterns.empty?
          nil
        else
          pattern = patterns.inject(&:|)
          case @color
          when bord.first_move_color
            bord.black ^= (bit_string | pattern)
            bord.white ^= pattern
          when bord.passive_move_color
            bord.white ^= (bit_string | pattern)
            bord.black ^= pattern
          else
            nil
          end
        end
      end

      def count(bord)
        count = 0
        case @color
        when bord.first_move_color
          bord = bord.black
        when bord.passive_move_color
          bord = bord.white
        else
          nil
        end
        64.times do |i|
          count += 1 if bord[i] == 1
        end
        count
      end

      private
      def convert_to_number(coordinate)
        bit_string = 0
        if coordinate =~ /[1-8][A-H]|[A-H][1-8]/i
          row, col = coordinate.slice(/[1-8][A-H]|[A-H][1-8]/i).split("").partition{|item| item.match(/[1-8]/)}
          bit_string = 1 << (REPLACE[col.first.downcase] + (8 * (row.first.to_i - 1)))
        else
        end
        bit_string
      end
    end
  end
  class Log
    def initialize(**options)
    end

    def info
    end

    def load
    end
  end
end
